from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np
from PIL import Image, ImageDraw, ImageFont

from highlighter.core.exceptions import require_package

try:
    import cv2
except ModuleNotFoundError as _:
    cv2 = None

from highlighter.core.const import OBJECT_CLASS_ATTRIBUTE_UUID

__all__ = [
    "draw_text",
    "draw_annotated_box",
    "draw_annotated_boxes",
    "draw_masks_on_image",
    "draw_polys_on_image",
    "draw_annotations_on_image",
]


def _to_pil_image_as_needed(pil_or_np_image):
    if isinstance(pil_or_np_image, np.ndarray):
        image_copy = Image.fromarray(pil_or_np_image)
    elif isinstance(pil_or_np_image, Image.Image):
        image_copy = pil_or_np_image.copy()
    else:
        raise ValueError(f"Invalide image type: {type(pil_or_np_image)}")
    return image_copy


def draw_annotations_on_image(pil_or_np_image, annotations):

    boxes = []
    labels = []
    confidences = []
    for anno in annotations:

        boxes.append(anno.location.bounds)
        object_class_obs = anno.get_observation(OBJECT_CLASS_ATTRIBUTE_UUID)
        if object_class_obs is not None:
            labels.append(anno.get_observation(OBJECT_CLASS_ATTRIBUTE_UUID).value.short_str())
        else:
            labels.append("")
        confidences.append(anno.datum_source.confidence)

    image_copy = _to_pil_image_as_needed(pil_or_np_image)
    return draw_annotated_boxes(image_copy, labels, boxes, confidences=confidences)


def draw_polys_on_image(pil_or_np_image, polys, labels: Optional[List[Optional[str]]] = None):
    """

    Args:
        polys: [[(x0,y0), ... (xn,yn)], ...]
        labels: [str, None, str, str] or None
    """
    if labels is None:
        labels = [None for _ in polys]

    assert len(labels) == len(polys)

    image_copy = _to_pil_image_as_needed(pil_or_np_image)

    # Create a drawing context
    draw = ImageDraw.Draw(image_copy)

    # Define the box color (you can change this if you prefer a different color)
    box_color = "red"  # Change this to the color of your choice

    # Draw rectangles for each box
    for poly, label in zip(polys, labels):
        try:
            draw.polygon(poly, outline=box_color, width=1)
        except ValueError as e:
            raise ValueError(f"{e}. Check each coordinate is a tuple, got {poly}")

    return image_copy


@require_package(cv2, "cv2", "opencv")
def draw_masks_on_image(image, masks, labels=None):
    """
    Draw boolean masks on the image with 10% transparency.

    Args:
        image (numpy.ndarray): Input image.
        masks (list of numpy.ndarray): List of boolean masks to draw.

    Returns:
        numpy.ndarray: Image with masks drawn.
    """
    # Convert image to RGBA format
    image_rgba = cv2.cvtColor(image, cv2.COLOR_RGB2BGRA)

    # Set transparency level (alpha channel) to 10%
    alpha = int(0.1 * 255)

    # Iterate through masks
    for mask in masks:
        # Convert boolean mask to uint8
        mask_uint8 = mask.astype(np.uint8) * 255

        # Create RGBA mask
        mask_rgba = cv2.cvtColor(mask_uint8, cv2.COLOR_GRAY2BGRA)

        # Set transparency level
        mask_rgba[:, :, 3] = alpha

        # Combine mask with image
        image_rgba = cv2.addWeighted(image_rgba, 1, mask_rgba, 1, 0)

    image_rgb = cv2.cvtColor(image_rgba, cv2.COLOR_RGBA2RGB)

    return image_rgb


BOX_XYXY = Tuple[int, int, int, int]
FONT = str(Path(__file__).parent / "DejaVuSans.ttf")
LINE_WEIGHT_SCALE = 0.004  # Just eyeballed this magic number
RGB = Tuple[int, int, int]
RGB_WHITE = (255, 255, 255)
TEXT_SIZE_SCALE = 0.02  # Just eyeballed this magic number
XY = Tuple[int, int]

label_colourer = None
label_to_color = {}


def draw_annotated_boxes(
    pil_img: Image.Image,
    labels: List[str],
    bboxes: List[BOX_XYXY],
    color_map: Optional[Dict[str, RGB]] = None,
    confidences: Optional[List[float]] = None,
):
    # print(f"boxes: {bboxes}")
    """Draw a list of boxes and labels onto an image"""
    global label_to_color, label_colourer

    if (color_map is None) and (label_colourer is None):
        # If no color_map is provided we instantiate the
        # label_colourer
        label_colourer = LabelColourer(as_int=True)
    elif color_map is not None:
        label_to_color = color_map

    if confidences is None:
        confidences = [None] * len(labels)

    for bbox, label, conf in zip(bboxes, labels, confidences):
        if label in label_to_color:
            color = label_to_color[label]
        elif color_map is not None:
            # Color map is set but a label cannot be found
            raise ValueError(f"'{label}' is not in color_map: '{color_map}'")
        else:
            color = label_colourer[label]
            label_to_color[label] = color

        pil_img = draw_annotated_box(pil_img, label, bbox, confidence=conf, color=color)
        # print(f"{label} - {bbox} - {conf} - {color}")
    return pil_img


def draw_annotated_box(
    pil_img: Image.Image,
    label: str,
    bbox: BOX_XYXY,
    confidence: Optional[float] = None,
    color: RGB = (209, 239, 8),
):
    image_height = pil_img.size[1]
    draw = ImageDraw.Draw(pil_img)
    # font = ImageFont.load_default()
    draw.rectangle(
        (int(bbox[0]), int(bbox[1]), int(bbox[2]), int(bbox[3])),
        width=int(LINE_WEIGHT_SCALE * image_height),
        outline=color,
    )

    if confidence is not None:
        label = f"{label}: {confidence:0.2f}"

    pil_img = draw_text(
        pil_img,
        label,
        (bbox[0], bbox[1]),
        text_color=color,
        bg_color=(0, 0, 0),
    )

    return pil_img


def draw_text(
    pil_img,
    text: str,
    bottom_left: XY,
    text_color: RGB = RGB_WHITE,
    bg_color: Optional[RGB] = None,
):
    image_height = pil_img.size[1]
    draw = ImageDraw.Draw(pil_img)

    fontsize = 1
    font = ImageFont.truetype(FONT, fontsize)

    # Fn the get the full bbox of a string including multiline strings
    def get_text_wh(text, font):
        text_height = 0
        text_width = []
        for line in text.split("\n"):
            # Add \n back in because it effects the height of the textbox
            # by accounting for the space between the lines
            _line = f"{line}\n"
            text_height += font.getbbox(_line)[3] - font.getbbox(_line)[1]
            text_width.append(font.getbbox(_line)[2] - font.getbbox(_line)[0])

        return max(text_width), text_height

    # Loop over some text to get a scaled text height
    # we're only interested in the hight of a single line
    # of text so we pass a dummy string as opposed to the
    # actual text argument
    while get_text_wh("dummy text", font)[1] < (image_height * TEXT_SIZE_SCALE):
        fontsize += 1
        font = ImageFont.truetype(FONT, fontsize)

    # Here we're interested in the bbox containing the whole
    # text so we pass the text argument not a dummpy string
    text_w, text_h = get_text_wh(text, font)

    offset = max(int(text_h * 0.04), 2)

    # We draw the text on top of the bounding box.
    # This means the top_left of the bounding box is
    # the bottom left of the text box.
    #
    #               --------------
    #               |x0,y0       |x1,y0
    #   text_box    |            |
    #               |x0,y1       |x1,y1
    #               -----------------------
    #               |bottom_left          |
    #               |                     |
    #    bbox       |                     |
    #               |                     |
    #               |                     |
    #               -----------------------
    x0 = int(bottom_left[0])
    y1 = int(bottom_left[1])

    x1 = int(x0 + text_w + (2 * offset))
    y0 = int(y1 - text_h - (2 * offset))

    if bg_color is not None:
        draw.rectangle(
            (x0, y0, x1, y1),
            width=1,
            outline=bg_color,
            fill=bg_color,
        )

    draw.text(
        (int(x0 + offset), int(y0 + offset)),
        text,
        fill=text_color,
        font=font,
    )
    return pil_img


class _CMAP_Tab20:
    """Equivalent to matplotlib.cm.tab20(i)"""

    _cmap = [
        (
            0.12156862745098039,
            0.4666666666666667,
            0.7058823529411765,
            1.0,
        ),
        (
            0.6823529411764706,
            0.7803921568627451,
            0.9098039215686274,
            1.0,
        ),
        (1.0, 0.4980392156862745, 0.054901960784313725, 1.0),
        (1.0, 0.7333333333333333, 0.47058823529411764, 1.0),
        (
            0.17254901960784313,
            0.6274509803921569,
            0.17254901960784313,
            1.0,
        ),
        (
            0.596078431372549,
            0.8745098039215686,
            0.5411764705882353,
            1.0,
        ),
        (
            0.8392156862745098,
            0.15294117647058825,
            0.1568627450980392,
            1.0,
        ),
        (1.0, 0.596078431372549, 0.5882352941176471, 1.0),
        (
            0.5803921568627451,
            0.403921568627451,
            0.7411764705882353,
            1.0,
        ),
        (
            0.7725490196078432,
            0.6901960784313725,
            0.8352941176470589,
            1.0,
        ),
        (
            0.5490196078431373,
            0.33725490196078434,
            0.29411764705882354,
            1.0,
        ),
        (
            0.7686274509803922,
            0.611764705882353,
            0.5803921568627451,
            1.0,
        ),
        (
            0.8901960784313725,
            0.4666666666666667,
            0.7607843137254902,
            1.0,
        ),
        (
            0.9686274509803922,
            0.7137254901960784,
            0.8235294117647058,
            1.0,
        ),
        (
            0.4980392156862745,
            0.4980392156862745,
            0.4980392156862745,
            1.0,
        ),
        (
            0.7803921568627451,
            0.7803921568627451,
            0.7803921568627451,
            1.0,
        ),
        (
            0.7372549019607844,
            0.7411764705882353,
            0.13333333333333333,
            1.0,
        ),
        (
            0.8588235294117647,
            0.8588235294117647,
            0.5529411764705883,
            1.0,
        ),
        (
            0.09019607843137255,
            0.7450980392156863,
            0.8117647058823529,
            1.0,
        ),
        (
            0.6196078431372549,
            0.8549019607843137,
            0.8980392156862745,
            1.0,
        ),
    ]
    N = len(_cmap)

    def __call__(self, i):
        return self._cmap[i]


_cmap_tab20 = _CMAP_Tab20()


def LabelColourer(cmap=None, as_int=False, consistent_by_key=True):  # noqa: N802
    """Returns a default dict where the default value is generated
    by the get_colour function.

    It simply iterates a counter and gets the next color from cmap.
    By default returns a float, but can also cast as_int.

    If consistent_by_key is True then the same key will produce the
    same colour each time the function is called. Even if it's called
    from a different process.
    """

    cmap = _cmap_tab20 if cmap is None else cmap

    # Get a color from cmap, convert to int rbg as needed
    # if idx is passed then use it otherwise update count and
    # use that number instead.
    def get_colour(idx=None):
        if idx is None:
            get_colour.count += 1
            *rgb, _ = cmap(get_colour.count % cmap.N)
        else:
            *rgb, _ = cmap(idx % cmap.N)
        if as_int:
            rgb = [int(255 * i) for i in rgb]
        return tuple(rgb)

    if consistent_by_key:

        def default_fatcory(key):
            # Sum the characters in the key by their position in
            # the ascii table
            idx = sum([ord(c) for c in key])

            # ToDo: check if there is a color clash, if there are remaining
            # colors increment by one untill you find a unique one.

            return get_colour(idx=idx)

        class KeyDependantDefaultDict(defaultdict):
            def __missing__(self, key):
                return default_fatcory(str(key))

        dd = KeyDependantDefaultDict(default_fatcory)
    else:
        get_colour.count = -1
        dd = defaultdict(get_colour)

    return dd



# To Do
# ~~~~~
# - In "sample_image()" replace "self.image" with an "image_queue"
#   Currently, will drop "images" if get_image() isn't called often enough
#   - Limit the queue length (ring buffer) to prevent excessive memory use
#   - Optionally, offer an "image queue", so that no images are lost
#
# - When pipeline can't be set to "playing state", find out more about the
#   GStreamer problem and report as part of the diagnostic message
#   - Improve command line utilities to catch and report the problem

import numpy as np
import sys
from threading import Thread
import time

if sys.version_info >= (3,0):
  from queue import Queue, Empty
else:
  from Queue import Queue, Empty

from aiko_services.elements.gstreamer import *

__all__ = ["VideoReader"]

# -----------------------------------------------------------------------------

Gst = gst_initialise()

class VideoReader:
  def __init__(self, pipeline, sink):
    self.pipeline  = pipeline
    self.bus       = pipeline.get_bus()
    self.queue     = Queue(maxsize=30)
    self.frame_id  = 1
    self.image     = None
    self.terminate = False
    self.timestamp = 0  # Unix time

#   sink = pipeline.get_by_name("sink")
#   sink.set_property("max-buffers", 2)
#   sink.set_property("drop", True)
#   sink.set_property("sync", False)
    sink.set_property("emit-signals", True)
    sink.connect("new-sample", self.sample_image, sink)

    if pipeline.set_state(Gst.State.PLAYING) == Gst.StateChangeReturn.FAILURE:
      raise GStreamerError("Unable to set the pipeline to the playing state")

    def _run():
      while not self.terminate:
        timeout = 1000 * 1000  # nanoseconds
        message = self.bus.timed_pop_filtered(timeout, Gst.MessageType.ANY)

        if self.image is not None:
#         print("NBVR frame: ", self.frame_id, self.queue.qsize())
#         self.image = cv2.cvtColor(self.image, cv2.COLOR_BGR2RGB)
          frame = {
            "type": "image",
            "id": self.frame_id,
            "image": self.image,
            "timestamp": self.timestamp  # Unix time
          }
          self.image = None
          self.timestamp = None
          self.queue.put(frame)
          self.frame_id += 1

        if message:
          if message.type == Gst.MessageType.ERROR:
            err, debug = message.parse_error()
            print("Error received from element %s: %s" % (
              message.src.get_name(), err))
            print("Debugging information: %s" % debug)
            break
          elif message.type == Gst.MessageType.EOS:
            print("GStreamer End-Of-Stream")
            frame = { "type": "EOS" }
            self.queue.put(frame)
            break
          elif message.type == Gst.MessageType.STATE_CHANGED:
            if isinstance(message.src, Gst.Pipeline):
              old_state,new_state,pending_state = message.parse_state_changed()
              print("GStreamer Pipeline state changed from %s to %s" %
                (old_state.value_nick, new_state.value_nick))

    self._t = Thread(target=_run, args=())
    self._t.daemon = True
    self._t.start()

  def gst_to_opencv(self, buffer_data, caps):
#   format = caps.get_structure(0).get_value("format")
    width  = caps.get_structure(0).get_value("width")
    height = caps.get_structure(0).get_value("height")
    image = np.ndarray((height, width, 3), buffer=buffer_data, dtype=np.uint8)
    return image

  def read_frame(self, timeout=None):
    try:
      return self.queue.get(block = timeout is not None, timeout = timeout)
    except Empty:
      return None

  def queue_size(self):
    return self.queue.qsize()

  def sample_image(self, sink, data):
    sample = sink.emit("pull-sample")
    self.timestamp = time.time()  # Unix time
    buffer = sample.get_buffer()

  # Avoid memory leak that is common in many on-line Python examples 😱
  # Use "buffer.map" instead of "buffer.extract_dup(0, buffer_size)" 🤔
    success, buffer_map = buffer.map(Gst.MapFlags.READ)

    if success:
      try:
        caps = sample.get_caps()
        self.image = self.gst_to_opencv(buffer_map.data, caps)

      # https://gstreamer.freedesktop.org/documentation/gstreamer/...
      #     gstbuffer.html#members
      # self.decode_timestamp = buffer.dts
      # self.presentation_timestamp = buffer.pts

      finally:
        buffer.unmap(buffer_map)
    return Gst.FlowReturn.OK

  def stop(self):
    self.terminate = True



import json
import os
from pathlib import Path
from typing import Dict

import click

from ..client import HLJSONEncoder, create_data_files, multithread_graphql_file_download
from ..core import LOG
from .common import _to_pathlib


@click.group("data-file")
@click.pass_context
def data_file_group(ctx):
    pass


@data_file_group.command("read")
@click.option(
    "-i",
    "--ids",
    type=int,
    required=False,
    multiple=True,
    default=[],
    show_default=True,
)
@click.option(
    "-p",
    "--paths",
    type=str,
    required=False,
    multiple=True,
    default=[],
    show_default=True,
)
@click.option(
    "-o",
    "--data-file-dir",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="Where to save the data_files to",
)
@click.pass_context
def read(ctx, ids, paths, data_file_dir):
    client = ctx.obj["client"]
    data_file_ids = list(ids)

    for path in paths:
        with open(path, "r") as f:
            lines = f.readlines()
        data_file_ids.extend([data_file_id.strip() for data_file_id in lines])

    result = multithread_graphql_file_download(
        client,
        data_file_ids,
        data_file_dir,
    )


@data_file_group.command("create")
@click.option(
    "-i",
    "--data-source-id",
    type=int,
    required=True,
)
@click.option(
    "-d",
    "--data-file-dir",
    type=click.Path(exists=True, file_okay=False),
    required=False,
    default="",
    help="Directory of data_files to upload",
    callback=_to_pathlib,
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    default=None,
    help=".txt file or coco.json",
    callback=_to_pathlib,
)
@click.option(
    "-m",
    "--multipart-filesize",
    type=str,
    required=False,
)
@click.pass_context
def create(ctx, data_source_id, data_file_dir, file, multipart_filesize):
    """Upload a collection of data_files to a Datasource

    \b
    Options:
        data_file_dir: A directory of data_files to upload, if 'file' is passed
        this will serve as the root directory for the data_file file names.
        If None this it is assumed the full path is available within 'file'

        file:
            Option 1 (None): No file is passed, we simply try to upload all files
            in `data_file_dir`. After the upload a directory at <data_file_dir>_with_hl_ids
            will be created containing simlinks mapping the newly created Highlighter
            data_file ids to the original paths.

            Option 2 (.txt): A text file with each line a path to a data_file
            to upload. After the upload a directory at <file.name>_with_hl_ids
            will be created containing simlinks mapping the newly created Highlighter
            data_file ids to the original paths.

            Option 3 (.json): A valid coco detection/segmentation dataset. If
            `data_file_dir` is used it will be prepended to all `images.file_name`
            fields. After the upload a directory at <file.name>_with_hl_ids
            will be created containing an `images` directory of symlinks mapping
            the newly created Highlighter data_file ids to the original paths and
            a copy of the original coco json file with the updated file_path
            pointing to the image file symlinks
    """
    client = ctx.obj["client"]

    class CocoFilePathSource:
        def __init__(self, coco_json: Path, file_dir: Path):
            self.coco_json = coco_json
            self.file_dir = file_dir
            self.output_root = self.coco_json.parent / f"{self.coco_json.stem}_with_hl_ids"
            self.output_json = self.output_root / self.coco_json.name
            self.output_images_dir = self.output_root / "images"
            self.output_images_dir.mkdir(parents=True, exist_ok=True)

            with Path(file).open("r") as f:
                self.coco_data = json.load(f)
                self.file_paths = [self.file_dir / i["file_name"] for i in self.coco_data["images"]]
                assert all([p.exists() for p in self.file_paths])

        def get_data_file_paths(self):
            return self.file_paths

        def update_data_file_ids(self, file_path_to_id: Dict[str, str]):
            old_id_to_new_id = {}
            for image_record in self.coco_data["images"]:
                original_file_path = self.file_dir / image_record["file_name"]
                new_id = file_path_to_id[str(original_file_path)]
                old_id_to_new_id[image_record["id"]] = new_id
                image_record["id"] = new_id

                link_path = (self.output_images_dir / f"{new_id}{original_file_path.suffix}").absolute()

                if link_path.exists():
                    os.unlink(link_path)

                link_path.symlink_to(original_file_path.absolute())

                image_record["file_name"] = str(link_path.name)

            for annotation_record in self.coco_data["annotations"]:
                old_image_id = annotation_record["image_id"]
                annotation_record["image_id"] = old_id_to_new_id[old_image_id]

            with self.output_json.open("w") as f:
                json.dump(self.coco_data, f)

        def message(self):
            return f"Duplicate of {self.coco_json} created with Highlighter file_ids created at {self.output_root}"

    class TextFilePathSource:
        def __init__(self, text_file_path: Path, file_dir: Path):
            self.text_file_path = text_file_path
            self.file_dir = file_dir
            self.output_dir = self.text_file_path.parent / f"{self.text_file_path.stem}_with_hl_ids"

            with Path(file).open("r") as f:
                self.file_paths = [self.file_dir / line.strip() for line in f.readlines()]
                assert all([p.exists() for p in self.file_paths])

        def get_data_file_paths(self):
            return self.file_paths

        def update_data_file_ids(self, file_path_to_id: Dict[str, str]):
            # Create the new directory if it doesn't exist
            self.output_dir.mkdir(exist_ok=True)

            # Iterate over the original and new filenames
            for original_path, data_file_id in file_path_to_id.items():
                # Create the full paths to the original and new files
                original_path = Path(original_path).absolute()
                link_path = (self.output_dir / f"{data_file_id}{original_path.suffix}").absolute()
                if link_path.exists():
                    os.unlink(link_path)

                # Check if the original file exists
                # Create a symlink to the original data_file with the new filename
                link_path.symlink_to(original_path)

        def message(self):
            return f"Directory of symlinks with Highlighter file_ids created at {self.output_dir}"

    class DirFilePathSource:
        def __init__(self, file_dir: Path):
            self.file_dir = file_dir
            self.output_dir = Path(f"{self.file_dir}_with_hl_ids")

        def get_data_file_paths(self):
            def path_iterator():
                for p in os.scandir(str(self.file_dir)):
                    if p.is_file():
                        yield Path(p.path)

            return path_iterator()

        def update_data_file_ids(self, file_path_to_id: Dict[str, str]):
            # Create the new directory if it doesn't exist
            self.output_dir.mkdir(exist_ok=True)

            # Iterate over the original and new filenames
            for original_path, data_file_id in file_path_to_id.items():
                # Create the full paths to the original and new files
                original_path = Path(original_path).absolute()
                link_path = (self.output_dir / f"{data_file_id}{original_path.suffix}").absolute()
                if link_path.exists():
                    os.unlink(link_path)

                # Check if the original file exists
                # Create a symlink to the original data_file with the new filename
                link_path.symlink_to(original_path)

        def message(self):
            return f"Directory of symlinks with Highlighter file_ids created at {self.output_dir}"

    if file is None:
        # If no file passed but we have data_file_dir then
        # we must be uploading a dir of data_files
        file_path_src = DirFilePathSource(data_file_dir)

    elif Path(file).suffix == ".json":
        # Json must be a coco file
        file_path_src = CocoFilePathSource(file, data_file_dir)

    else:
        # If not json it must be a text file
        file_path_src = TextFilePathSource(file, data_file_dir)

    data_file_path_to_id, failed_data_file_paths = create_data_files(
        client,
        file_path_src.get_data_file_paths(),
        data_source_id,
        progress=True,
        multipart_filesize=multipart_filesize,
    )

    file_path_src.update_data_file_ids(data_file_path_to_id)

    LOG.info(
        f"{len(data_file_path_to_id)} succeeded, {len(failed_data_file_paths)} failed -> {failed_data_file_paths}"
    )
    LOG.info(file_path_src.message())

    click.echo(
        json.dumps(
            {
                "data_file_path_to_id": data_file_path_to_id,
                "failed_data_file_paths": failed_data_file_paths,
            },
            cls=HLJSONEncoder,
        )
    )



from pathlib import Path

import click
import yaml
from pydantic import BaseModel

from highlighter.cli.common import CommonOptions

from ..client import get_latest_assessments_gen, multithread_graphql_file_download
from ..core import DEFAULT_PAGE_SIZE
from ..datasets import (
    SUPPORTED_SPLIT_FNS,
    CocoWriter,
    Dataset,
    DatasetFormat,
    YoloWriter,
    get_reader,
    get_split_fn,
)


def _download_data_files(client, data_file_ids, data_file_dir):
    if data_file_dir is not None:
        multithread_graphql_file_download(
            client,
            data_file_ids,
            data_file_dir,
        )


def _combine_datasets(ctx):
    client = ctx.obj["client"]
    page_size = ctx.obj["page_size"]
    dataset_splits = ctx.obj["dataset_splits"]

    downloaded = {}
    datasets = []
    for dataset_id, split in dataset_splits:
        if dataset_id in downloaded:
            ds = Dataset(
                annotations_df=downloaded[dataset_id].annotations_df.copy(),
                data_files_df=downloaded[dataset_id].data_files_df.copy(),
            )
        else:
            ds = Dataset.read_from(
                dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
                client=client,
                dataset_id=dataset_id,
                page_size=page_size,
            )
        ds.data_files_df.split = split
        datasets.append(ds)
        downloaded[dataset_id] = ds

    return Dataset.combine(datasets)


@click.group("dataset")
@click.pass_context
def dataset_group(ctx):
    pass


@dataset_group.command(
    "create",
    help="""
Create a Dataset in Highlighter

\b
Note: If using the --dataset-id param. There are no safety rails! That is to
say; the cli will allow duplicate assessments and data_files if you ask it to.

  ie: If dataset 123 has data_file 456 and workflow 789 also has data_file 456 the
  assessments for both will be added to the dataset.

Typically you would use the --dataset-id param if you wanted to combine 2 or
more datasets, or combine some new data from a workflow with an existing dataset.
In which case a new dataset would be created that simply has all assessments
appended together.
        """,
)
@click.option(
    "-n",
    "--name",
    type=str,
    required=True,
)
@click.option(
    "-pid",
    "--workflow-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-oid",
    "--object-class-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-uid",
    "--user-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-did",
    "--dataset-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
    help=(
        "Add assessments from existing dataset(s), see Note in help string "
        " for more context and a warrning."
    ),
)
@click.option(
    "--split-type",
    type=click.Choice(SUPPORTED_SPLIT_FNS.keys()),
    required=False,
    default="RandomSplitter",
    show_default=True,
)
@click.option(
    "--split-seed",
    type=int,
    required=False,
    default=42,
    show_default=True,
)
@click.option(
    "--split-frac",
    type=click.Tuple([str, float]),
    required=False,
    multiple=True,
    default=[("data", 1.0)],
    show_default=True,
)
@click.option(
    "-c",
    "--classes-of-interest",
    type=str,
    required=False,
    default=None,
    multiple=True,
    help="Object class uuids to consider when validating splits",
)
@click.pass_context
def create(
    ctx,
    name,
    workflow_id,
    object_class_id,
    user_id,
    dataset_id,
    split_type,
    split_seed,
    split_frac,
    classes_of_interest,
):
    client = ctx.obj["client"]

    query_args = dict(
        projectId=workflow_id,
        objectClassId=object_class_id,
        userId=user_id,
    )

    def not_empty(v):
        if v is None:
            return False
        if v == ():
            return False
        if v == []:
            return False
        return True

    query_args = {k: v for k, v in query_args.items() if not_empty(v)}

    datasets = []
    if query_args:
        latest_subs_gen = get_latest_assessments_gen(
            client,
            **query_args,
        )

        reader = get_reader("highlighter_assessments")(latest_subs_gen)
        ds = Dataset.load_from_reader(reader)
        datasets.append(ds)

    dataset_description_fields = []
    if dataset_id:
        # add to query_args here because datasetId is not a valid
        # arg to get_latest_assessments_gen, but we still want it
        # to appear in the Highlighter Dataset description.
        query_args["datasetId"] = dataset_id
        dataset_list = [
            Dataset.read_from(
                dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
                client=client,
                dataset_id=ds_id,
                page_size=DEFAULT_PAGE_SIZE,
            )
            for ds_id in dataset_id
        ]

        datasets.extend(dataset_list)

        def get_dataset_url(id, client=client):
            return client.endpoint_url.replace("graphql", f"datasets/{id}")

        def get_dataset_name(id, client=client):
            class DatasetNameOnly(BaseModel):
                name: str

            name = client.dataset(
                return_type=DatasetNameOnly,
                id=id,
            ).name
            return name.replace("_", "\\_")

        base_dataset_links = [
            f"  - [{id}]({get_dataset_url(id)}): {get_dataset_name(id)}" for id in dataset_id
        ]
        base_dataset_links_str = "\n".join(base_dataset_links)
        dataset_description_fields.append(("Base Datasets", base_dataset_links_str))

    dataset = Dataset.combine(datasets)
    if classes_of_interest:
        adf = dataset.annotations_df
        ddf = dataset.data_files_df

        adf = adf[adf.value.isin(classes_of_interest)]
        ddf = ddf[ddf.data_file_id.isin(adf.data_file_id.unique())]
        dataset.annotations_df = adf
        dataset.data_files_df = ddf

    split_names = [s[0] for s in split_frac]
    fracs = [s[1] for s in split_frac]
    splitter = get_split_fn(split_type)(split_seed, fracs, split_names)

    dataset.apply_split(splitter)

    split_args = {
        "type": splitter.__class__.__name__,
        "seed": splitter.seed,
        "fracs": splitter.fracs,
        "names": splitter.names,
    }

    split_str = yaml.safe_dump(split_args)
    split_str = f"<pre>" + split_str + "</pre> \n"

    query_str = yaml.safe_dump(query_args)
    query_str = f"<pre>" + query_str + "</pre> \n"

    # List of tuples with e[0] being the heading and e[1]
    # being the content. This will be rendered as Markdown
    # in the publish_to_highlighter function.
    dataset_description_fields.extend(
        [
            ("Query", query_str),
            ("Split", split_str),
        ]
    )

    dataset.publish_to_highlighter(
        client,
        name,
        dataset_description_fields=dataset_description_fields,
        split_fracs={s: f for s, f in zip(split_names, fracs)},
    )


@click.group("read")
@click.option(
    "-i",
    "--dataset-ids",
    type=str,
    required=False,
    multiple=True,
    default=[],
    show_default=True,
    help=" int or int:str --> ID or ID:SPLIT",
)
@click.option(
    "--page-size",
    type=int,
    required=False,
    default=DEFAULT_PAGE_SIZE,
    show_default=True,
)
@click.pass_context
def read_group(ctx, dataset_ids, page_size):
    """Read Datasets from Highlighter and write to local disk.

    For help with each specific output format use:

    \b
      hl dataset read coco|yolo|... --help

    \b
    By default Datasets spits are all set to 'data' to assign
    a split to a dataset use --dataset-id ID:SPLIT

    \b
    Example:
      read datasets 111 and 222, assign 111 to 'train' split and
      222 to 'val' split then write to yolo detection format

    \b
      hl dataset read -i 111:train -i 204:val yolo -t detection my_output
    """
    ctx.obj["page_size"] = page_size

    if all([":" in i for i in dataset_ids]):
        dataset_splits = [i.split(":") for i in dataset_ids]
        dataset_splits = [(int(id), split) for id, split in dataset_splits]
    else:
        dataset_splits = [(int(i), "data") for i in dataset_ids]
    ctx.obj["dataset_splits"] = dataset_splits


@read_group.command("hdf")
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def write_hdf(ctx, annotations_dir, data_file_dir):
    filestem = "-".join([str(i[0]) for i in ctx.obj["dataset_splits"]])
    filename = f"{filestem}.hdf"

    hdf_path = annotations_dir / filename
    combined_dataset = _combine_datasets(ctx)
    combined_dataset.write_hdf(hdf_path)

    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


@read_group.command("json")
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def write_json(ctx, annotations_dir, data_file_dir):
    filestem = "-".join([str(i[0]) for i in ctx.obj["dataset_splits"]])
    filename = f"{filestem}.json"

    json_path = annotations_dir / filename
    combined_dataset = _combine_datasets(ctx)
    combined_dataset.write_json(json_path)

    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


@read_group.command(CocoWriter.format_name)
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def write_coco(ctx, annotations_dir, data_file_dir):
    writer = CocoWriter(annotations_dir)

    combined_dataset = _combine_datasets(ctx)

    writer.write(combined_dataset)
    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


@read_group.command(YoloWriter.format_name)
@click.argument("output_dir", type=str)
@CommonOptions.data_file_dir
@click.option(
    "--task",
    "-t",
    type=click.Choice([member.value for member in YoloWriter.TASK]),
    default=YoloWriter.TASK.DETECT,
)
@click.pass_context
def write_yolov8(ctx, output_dir, data_file_dir, task):
    image_cache_dir = Path(".") if data_file_dir is None else data_file_dir
    writer = YoloWriter(output_dir, image_cache_dir, task=task)

    combined_dataset = _combine_datasets(ctx)
    writer.write(combined_dataset)

    data_file_ids = list(combined_dataset.data_files_df.data_file_id.unique())
    client = ctx.obj["client"]
    _download_data_files(client, data_file_ids, data_file_dir)


dataset_group.add_command(read_group)



from abc import ABC, abstractmethod
from typing import Tuple

import pandas as pd

from ..dataset import Dataset

__all__ = ["DatasetSplitter"]


class DatasetSplitter(ABC):
    @abstractmethod
    def split(self, dataset: Dataset) -> Tuple[pd.DataFrame, pd.DataFrame]:
        pass



from uuid import UUID, uuid4
from warnings import warn


def _get_uuid_str():
    return str(uuid4())


def _validate_uuid(v):
    if isinstance(v, UUID):
        return str(v)

    if v is None:
        warn("entity_id was not provided, generating one")
        return _get_uuid_str()

    try:
        _ = UUID(v)
        return v
    except:  # noqa
        raise ValueError("Invalid UUID string")



import asyncio
import concurrent.futures
import os
import time
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Union

import boto3
import yaml
from aiohttp.client_exceptions import ClientConnectorError
from gql import Client as GQLClient
from gql import gql
from gql.transport.aiohttp import AIOHTTPTransport
from pydantic import BaseModel

from highlighter.client._colors import ColoredString

from ..core import get_default_logger, snake_to_camel
from .gql_base import (
    Line,
    get_all_mutations,
    get_all_queries,
    get_gql_obj,
    get_gql_return_type,
    get_gql_schema,
    return_type_formatting,
    to_gql_type,
    to_python_type,
)

LOGGER = get_default_logger("HLClient")

cs = ColoredString()


KEY_API_TOKEN = "api_token"  # nosec hardcoded_password_string
KEY_ENDPOINT_URL = "endpoint_url"
KEY_CLOUD = "cloud"

KEY_AWS_ACCESS_KEY_ID = "aws_access_key_id"
KEY_AWS_SECRET_ACCESS_KEY = "aws_secret_access_key"  # nosec hardcoded_password_string
KEY_AWS_REGION = "aws_default_region"

ENV_HL_WEB_GRAPHQL_API_TOKEN = "HL_WEB_GRAPHQL_API_TOKEN"  # nosec hardcoded_password_string
ENV_HL_WEB_GRAPHQL_ENDPOINT = "HL_WEB_GRAPHQL_ENDPOINT"
ENV_AWS_ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
ENV_AWS_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"  # nosec hardcoded_password_string
ENV_AWS_DEFAULT_REGION = "AWS_DEFAULT_REGION"

ENV_HL_DEFAULT_PROFILE = "HL_DEFAULT_PROFILE"
ENV_HL_PROFILES_YAML = "HL_PROFILES_YAML"
ENV_HL_GQL_TIMEOUT_SEC = "HL_GQL_TIMEOUT_SEC"

CONST_HLCLIENT_GQL_TIMEOUT_SEC = os.environ.get(ENV_HL_GQL_TIMEOUT_SEC, 60)
CONST_GRAPHQL_DEFAULT_PROFILE = os.environ.get(ENV_HL_DEFAULT_PROFILE, None)
CONST_DEFAULT_GRAPHQL_PROFILES_YAML = Path.home() / ".highlighter-profiles.yaml"

CONST_GRAPHQL_PROFILES_YAML = os.environ.get(ENV_HL_PROFILES_YAML, CONST_DEFAULT_GRAPHQL_PROFILES_YAML)

EXAMPLE_PROFILE = {
    "my-first-profile": {
        KEY_ENDPOINT_URL: "https://<client-account>.highlighter.ai/graphql",
        KEY_API_TOKEN: "123...abc",
    }
}


class S3Creds(BaseModel):
    __type__: str = "aws-s3"
    type: Literal[__type__] = __type__
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_default_region: str

    def as_environment_variables(self) -> Dict[str, str]:
        return {
            ENV_AWS_ACCESS_KEY_ID: self.aws_access_key_id,
            ENV_AWS_SECRET_ACCESS_KEY: self.aws_secret_access_key,
            ENV_AWS_DEFAULT_REGION: self.aws_default_region,
        }


def validate_cloud_creds(cloud_creds_dict):
    valid_cloud_cred_types = [S3Creds]
    for t in valid_cloud_cred_types:
        try:
            creds = t(**cloud_creds_dict)
            return creds
        except:
            continue  # nosec try_except_continue
    raise ValueError(f"Not a valid cloud credential: {cloud_creds_dict}")


def get_credentials_from_profiles_yaml(
    profile,
    profiles_path=CONST_GRAPHQL_PROFILES_YAML,
):
    if not Path(profiles_path).exists():
        raise FileNotFoundError()

    with open(profiles_path, "r") as f:
        creds = yaml.safe_load(f).get(profile)

    if creds is None:
        raise KeyError(f"Profile '{profile}' could not be found in '{profiles_path}'")

    # Used to access private cloud storage outside of Highlighter S3
    cloud_cred_dicts = creds.get(KEY_CLOUD, None)
    if cloud_cred_dicts is not None:
        cloud_creds = {}
        for cred_dict in cloud_cred_dicts:
            _creds = validate_cloud_creds(cred_dict)
            cloud_creds[_creds.type] = _creds
    else:
        cloud_creds = None

    return creds[KEY_API_TOKEN], creds[KEY_ENDPOINT_URL], cloud_creds


def try_get_existing_asyncio_loop():
    try:
        loop = asyncio.get_running_loop()
    except:
        loop = None

    return loop


class HLClient(object):
    EXECUTION_RETRYS = 10

    # Set to True if running in a notebook or any environment what
    # is running an an async loop
    # Manually set to bool True | False if you wish to override manually
    _async: Union[str, bool] = "AUTO"
    _instance = None

    def __init__(self, client: GQLClient, cloud_creds: Optional[Dict[str, str]] = None):
        self._client = client
        self._schema = get_gql_schema(self)

        # Used to access private cloud storage outside of Highlighter S3
        self.cloud_creds = cloud_creds
        HLClient._instance = self

    @classmethod
    def from_profile(cls, profile: str, profiles_path=None) -> "HLClient":
        profiles_path = profiles_path or CONST_DEFAULT_GRAPHQL_PROFILES_YAML
        api_token, endpoint_url, cloud_creds = get_credentials_from_profiles_yaml(
            profile, profiles_path=profiles_path
        )

        return cls.from_credential(api_token, endpoint_url, cloud_creds=cloud_creds)

    @classmethod
    def from_credential(
        cls,
        api_token: str,
        endpoint_url: str,
        cloud_creds: Optional[Dict[str, Any]] = None,
    ) -> "HLClient":
        transport = AIOHTTPTransport(
            url=endpoint_url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Token {api_token}",
            },
        )
        client = GQLClient(
            transport=transport,
            fetch_schema_from_transport=True,
            execute_timeout=CONST_HLCLIENT_GQL_TIMEOUT_SEC,
        )
        return cls(client, cloud_creds=cloud_creds)

    def __repr__(self):
        return f"{self.endpoint_url}: [{self.api_token[:4]}...]"

    @classmethod
    def get_client(cls):
        """Instantiate a HLClient object as needed

        Singelton. If HLClient has already been instantiated this will return
        the existing instance, else will fallback on HLClient.get_client
        """
        if HLClient._instance is not None:
            return HLClient._instance
        return cls.from_env()

    @classmethod
    def from_env(cls) -> "HLClient":
        api_token = os.environ[ENV_HL_WEB_GRAPHQL_API_TOKEN]
        endpoint_url = os.environ[ENV_HL_WEB_GRAPHQL_ENDPOINT]
        cloud_creds = cls.cloud_creds_from_env()
        return cls.from_credential(api_token, endpoint_url, cloud_creds=cloud_creds)

    @classmethod
    def cloud_creds_from_env(cls) -> Optional[Dict[str, S3Creds]]:
        session = boto3.Session()
        session_credentials = session.get_credentials()
        aws_default_region = session.region_name
        aws_access_key_id = session_credentials.access_key if session_credentials is not None else None
        aws_secret_access_key = session_credentials.secret_key if session_credentials is not None else None

        if (
            (aws_default_region is not None)
            and (aws_access_key_id is not None)
            and (aws_secret_access_key is not None)
        ):
            cloud_creds = {
                S3Creds.__type__: S3Creds(
                    type=S3Creds.__type__,
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key,
                    aws_default_region=aws_default_region,
                )
            }
        else:
            cloud_creds = None
        return cloud_creds

    def export_credentials_to_environment(self):
        """Export credentials to environment variables."""
        os.environ[ENV_HL_WEB_GRAPHQL_API_TOKEN] = self.api_token
        os.environ[ENV_HL_WEB_GRAPHQL_ENDPOINT] = self.endpoint_url

        if self.cloud_creds is not None:
            for key, val in self.cloud_creds.items():
                if key == S3Creds.__type__:
                    for var_name, var in val.dict().items():
                        if var_name != "type":
                            os.environ[var_name.upper()] = var

    def append_credentials_to_env_file(self, outfile: str):
        """Append export KEY=VALUE lines for stored credentials"""
        token = self.api_token
        endpoint = self.endpoint_url

        # Token and Endpoint should always be present
        lines = [
            f"export {ENV_HL_WEB_GRAPHQL_API_TOKEN}={token}",
            f"export {ENV_HL_WEB_GRAPHQL_ENDPOINT}={endpoint}",
        ]

        # Cloud creds are optional. They are only needed if the client needs
        # to download data from a 3rd party's cloud bucket

        if self.cloud_creds is not None:
            for cloud_type, creds in self.cloud_creds.items():
                lines.extend(
                    [f"export {key}={value}" for key, value in creds.as_environment_variables().items()]
                )

        with open(str(outfile), "a+") as f:
            f.write("\n".join(lines))

    @property
    def endpoint_url(self):
        return self._client.transport.url

    @property
    def api_token(self):
        return self._client.transport.headers["Authorization"].split()[1]

    @property
    def account_name(self):
        return self.endpoint_url.split("//")[-1].split(".")[0]

    def get_s3_client(self):
        assert self.cloud_creds is not None
        assert S3Creds.__type__ in self.cloud_creds
        s3_creds = self.cloud_creds[S3Creds.__type__]
        s3_client = boto3.client(
            "s3",
            region_name=s3_creds.aws_default_region,
            aws_access_key_id=s3_creds.aws_access_key_id,
            aws_secret_access_key=s3_creds.aws_secret_access_key,
        )
        return s3_client

    def execute(self, request_str: str, variable_values=None) -> Dict[str, dict]:
        result = None
        error = None
        for i in range(self.EXECUTION_RETRYS):
            try:
                result = self._execute(request_str, variable_values=variable_values)
                break
            except ClientConnectorError as e:
                LOGGER.info(f"ClientConnectionError: {e} -- " f"Retrying {i+1}/{self.EXECUTION_RETRYS}")
                error = e
                time.sleep(1)

        if (result is None) and (error is not None):
            raise (error)
        return result

    def _execute(self, request_str: str, variable_values=None):
        LOGGER.debug(f"_execute: {request_str}")
        request_gql = gql(request_str)

        # Change behaviour if an asyncio loop is already running.
        # This is important when running in notebook enviroments.
        # I suggest keeping as "AUTO" but if for somereason you need
        # to maunually set this flag you can
        if self._async == "AUTO":
            loop = try_get_existing_asyncio_loop()
        elif self._async == True:
            loop = asyncio.get_running_loop()
        elif self._async == False:
            loop = None
        else:
            raise ValueError(
                (
                    f"Invalid value for HLClient._async expected one of: (True, False, 'AUTO'). "
                    f"Got {self._async}"
                )
            )

        if loop is None:
            result = self._client.execute(request_gql, variable_values=variable_values)
        else:
            # An event loop is already running (probs because this is being run
            # from Jupyter, Spyder or an IPython interpreter
            pool = concurrent.futures.ThreadPoolExecutor()
            result = pool.submit(
                asyncio.run,
                self._client.execute_async(
                    request_gql,
                    variable_values=variable_values,
                ),
            ).result()

        return result

    def __getstate__(self):
        state = dict(
            api_token=self.api_token,
            endpoint_url=self.endpoint_url,
            cloud_creds=self.cloud_creds,
        )
        return state

    def __setstate__(self, d):
        client = HLClient.from_credential(**d)
        self.__dict__ = client.__dict__.copy()

    def __getattr__(self, key) -> Callable[..., Any]:
        # methods = vars(type(self))
        # if key in methods:
        #     return methods[key]

        key = snake_to_camel(key)
        if key not in get_all_queries(self._schema) + get_all_mutations(self._schema):
            raise ValueError(f"{key} is not a known query or mutation")

        def f(*, return_type: BaseModel, **kwargs):
            return_type_dict = get_gql_return_type(return_type)

            obj_type, target_gql_obj = get_gql_obj(self._schema, key)

            arg_lst = []
            arg_names = [x["name"] for x in target_gql_obj["args"]]
            for k in kwargs:
                if k not in arg_names:
                    raise ValueError(f"unknown gql argument '{k}' for '{key}'")

            for x in target_gql_obj["args"]:
                if x["name"] in kwargs:
                    arg_lst.append((x["name"], to_gql_type(x["type"])))

            indent = 0
            if len(arg_lst):
                lines = [
                    Line(line="%s _(" % obj_type.lower(), indent=indent),
                    *[Line(line=f"${x}: {t}", indent=indent + 1) for x, t in arg_lst],
                    Line(line=")", indent=indent),
                    Line(line="{", indent=indent),
                    Line(line=f"{key}(", indent=indent + 1),
                    *[Line(line=f"{x}: ${x}", indent=indent + 2) for x, _ in arg_lst],
                    Line(line=f")", indent=indent + 1),
                    *return_type_formatting(return_type_dict, indent + 1),
                    Line(line="}", indent=indent),
                ]
            else:
                lines = [
                    Line(line="%s _" % obj_type.lower(), indent=indent),
                    Line(line="{", indent=indent),
                    Line(line=f"{key}", indent=indent + 1),
                    *return_type_formatting(return_type_dict, indent + 1),
                    Line(line="}", indent=indent),
                ]

            generated_request_str = "\n".join([" " * 4 * x.indent + x.line for x in lines])
            LOGGER.debug(f"executing {obj_type}:\n{generated_request_str}\nWith args {kwargs}")
            response = self.execute(generated_request_str, variable_values=kwargs)
            if "errors" in response:
                raise ValueError(response["errors"])

            result = response[key]
            try:
                return return_type(**result)
            except Exception as e:
                LOGGER.debug(f"response: {response}")
                if getattr(return_type, "_name", None) == "List":
                    # for List[BaseModel]
                    return [return_type.__args__[0](**x) for x in result]
                raise e

        return f

    def hint(self, query_or_mutation: str):
        obj_type, target_gql_obj = get_gql_obj(self._schema, query_or_mutation)
        py_pos_arg_lst = []
        py_key_arg_lst = []
        for x in target_gql_obj["args"]:
            py_type = to_python_type(x["type"])
            if not py_type.startswith("Optional"):
                py_pos_arg_lst.append((x["name"], py_type))
            else:
                py_key_arg_lst.append((x["name"], py_type))

        args = [
            (
                Line(line=f"{n}: {t} = None,", indent=0)
                if t.startswith("Optional")
                else Line(line=cs.yellow(f"{n}: {t},"), indent=0)
            )
            for n, t in py_pos_arg_lst + py_key_arg_lst
        ]
        print(cs.red_black(query_or_mutation))
        print("\n".join([" " * 4 * x.indent + x.line for x in args]))



import json
from pathlib import Path
from typing import Dict, List, Optional, Union

from shapely import wkt

from ....core import OBJECT_CLASS_ATTRIBUTE_UUID, PIXEL_LOCATION_ATTRIBUTE_UUID
from ...interfaces import IWriter
from .common import (
    CocoAnnotationRecord,
    CocoCategory,
    CocoImageRecord,
    get_bbox_area,
    segmentation_to_bbox,
    shapely_to_segmentation,
)

PathLike = Union[str, Path]


def validate_categories(
    categories: List[CocoCategory],
    unique_object_classes,
):
    # Validate no duplicate ids
    unique_ids = set([c.id for c in categories])
    if len(unique_ids) != len(categories):
        raise ValueError(f"Got dupliate category ids in: {categories}")

    # Validate ids are consecutive an start at 1
    if not all([a == b for a, b in zip(sorted(unique_ids), range(0, len(unique_ids)))]):
        raise ValueError("category.id are not consecutive and starting at 1 got: " f"{categories}")

    # Validate all unique_object_classes are accounted for in
    # categories
    category_names = [c.name for c in categories]
    if (len(category_names) != len(unique_object_classes)) or (
        len(set(category_names) - set(unique_object_classes)) > 0
    ):
        diff_table = _tabulate_list_diff(
            sorted(category_names), sorted(unique_object_classes), "coco_categories", "hl dataset classes"
        )
        raise ValueError(f"categories dont match unique_object_classes, got: \n{diff_table}")

    return categories


def _tabulate_list_diff(a, b, a_title, b_title):
    """Create a readable diff for 2 lists

    abc | abc
    xyz | ---
    --- | foo
    bar | bar

    """
    ab = sorted(set(a + b))
    left_cols = []
    right_cols = []
    for ele in ab:
        left_cols.append(ele if ele in a else "-" * len(ele))
        right_cols.append(ele if ele in b else "-" * len(ele))

    table = [f"{a_title:<40} | {b_title}"]
    table.extend([f"{l:40} | {r}" for l, r in zip(left_cols, right_cols)])
    return "\n".join(table)


class CocoWriter(IWriter):
    format_name = "coco"
    annotation_count = -1
    image_count = -1

    def __init__(
        self,
        annotations_dir: PathLike,
        category_attribute_id: str = OBJECT_CLASS_ATTRIBUTE_UUID,
        categories: Optional[List[Union[CocoCategory, Dict]]] = None,
    ):
        """Write annotations files.

        Args:
            annotations_dir: the dir to store the JSON file
            class_attribute_intentifier: value from the annotations_df.attribute_id  column select categories from
            categories: value from the annotations_df.values column to use as categories

            TODO: negative_sampling: if True, create bbox with (0, 0, 0, 0) in annotation for negative sampling
              creating 'pure' COCO JSON files.

        Example:
            | data_file_id | attribute_id | value |
            | ------------ | ------------ | ----- |
            | 001          | 1            | a     |
            | 001          | 1            | c     |
            | 001          | 2            | x     |
            | 002          | 1            | a     |
            | 002          | 1            | b     |
            | 002          | 2            | y     |

            If category_attribute_id="1" and categories=None
            the we would create a coco dataset with the categories [a,b,c]

            If category_attribute_id="1" and categories=[a,c]
            the we would create a coco dataset with the categories [a,c]

            If category_attribute_id="2" and categories=None
            the we would create a coco dataset with the categories [x,y]
        """
        self.annotations_dir = Path(annotations_dir)
        self.category_attribute_id = category_attribute_id

        def to_base_model(c):
            if isinstance(c, dict):
                return CocoCategory(**c)
            return c

        # Ensure we're working with the CocoCategory BaseModel
        if categories is None:
            self.categories = None
        else:
            self.categories: List[CocoCategory] = [to_base_model(c) for c in categories]

    def write(
        self,
        dataset: "Dataset",
    ):
        self.annotations_dir.mkdir(parents=True, exist_ok=True)

        unique_object_classes = dataset.annotations_df[
            dataset.annotations_df.attribute_id == self.category_attribute_id
        ].value.unique()

        # sort so datasets with the same object_classes
        # have identical 'categories' when saved to Coco format.
        # This assures when using different datasets (ie train and test)
        # that were download separately that the categories are alligned
        # so evaluation works. This can be overridden by using the optional
        # categories arg
        unique_object_classes.sort()

        # Allign categories with passed in categories
        # else fall back on sorted unique_object_classes
        if self.categories is not None:
            self.categories = validate_categories(self.categories, unique_object_classes)
            self.categories = sorted(self.categories, key=lambda c: c.id)
            self.object_class_value_to_cat_id = {}
            self.coco_categories = []
            for cat in self.categories:
                self.coco_categories.append(cat.dict())
                self.object_class_value_to_cat_id[cat.name] = cat.id

        else:
            self.object_class_value_to_cat_id = {
                object_class: i for i, object_class in enumerate(unique_object_classes)
            }

            self.coco_categories = [
                {"id": i, "name": object_class} for i, object_class in enumerate(unique_object_classes)
            ]

        def to_coco_annotation(grp):
            entity_id = grp.name
            if grp.shape[0] != grp.attribute_id.unique().shape[0]:
                print(f"WANING: DROPING DUPLICATE ROWS for entity: {grp.name}")
                grp = grp.drop_duplicates(subset=["attribute_id"])

            data = grp.set_index("attribute_id").to_dict("index")

            # Initially try both uuid and label
            # we'll probs just move to uuid I think
            object_class_attr = data.pop(
                OBJECT_CLASS_ATTRIBUTE_UUID,
                data.pop(str(OBJECT_CLASS_ATTRIBUTE_UUID), data.pop(OBJECT_CLASS_ATTRIBUTE_UUID.label, None)),
            )

            if object_class_attr is None:
                print((f"entity: {entity_id}, No OBJECT_CLASS"))
                # "expected an object class attribute for each entity "
                # f"got: {grp}"))
                return None

            object_class_value = object_class_attr["value"]
            category_id = self.object_class_value_to_cat_id[object_class_value]
            data_file_id = object_class_attr["data_file_id"]

            location_attr = data.pop(
                PIXEL_LOCATION_ATTRIBUTE_UUID,
                data.pop(
                    str(PIXEL_LOCATION_ATTRIBUTE_UUID), data.pop(PIXEL_LOCATION_ATTRIBUTE_UUID.label, None)
                ),
            )

            if location_attr is None:
                print((f"entity: {entity_id}, No LOCATION"))
                # "expected a location attribute for each entity "
                # f"got: {grp}"))
                return None

            location = location_attr["value"]
            if isinstance(location, str):
                location = wkt.loads(location)

            segmentation = shapely_to_segmentation(location)
            bbox = segmentation_to_bbox(segmentation)
            area = get_bbox_area(bbox)

            extra_fields = {attr_id: attr["value"] for attr_id, attr in data.items()}
            extra_fields["entity_id"] = str(entity_id)
            self.annotation_count += 1
            return CocoAnnotationRecord(
                id=self.annotation_count,
                image_id=data_file_id,
                category_id=category_id,
                segmentation=segmentation,
                area=area,
                bbox=bbox,
                # Added to schema to allow extensions and interoperability
                # with Highlighter AnnotationType.metadata
                extra_fields=extra_fields,
            ).dict()

        def to_coco_image(row):
            self.image_count += 1
            filename = Path(row.filename)
            stem, suffix = filename.stem, filename.suffix.lower()
            filename = f"{stem}{suffix}"
            return CocoImageRecord(
                id=row.data_file_id,
                file_name=filename,
                width=row.width,
                height=row.height,
                # Added to schema to allow extensions and interoperability
                # with Highlighter Image.metadata
                extra_fields=row.extra_fields,
                # Added so multiple splits can be held in one CocoDataset
                split=row.split,
            ).dict()

        for split in dataset.data_files_df.split.unique():
            split_data_files_df = dataset.data_files_df[dataset.data_files_df.split == split]
            split_annos_df = dataset.annotations_df[
                dataset.annotations_df.data_file_id.isin(split_data_files_df.data_file_id)
            ]
            coco_annotations = split_annos_df.groupby("entity_id").apply(to_coco_annotation)
            coco_annotations = [e for e in coco_annotations if e is not None]
            coco_images = split_data_files_df.apply(lambda row: to_coco_image(row), axis=1).to_list()
            coco = {
                "annotations": coco_annotations,
                "images": coco_images,
                "categories": self.coco_categories,
            }
            with (self.annotations_dir / f"{split}.json").open("w") as f:
                json.dump(coco, f)



import aiko_services as aiko

from highlighter.client import HLClient
from highlighter.client.tasks import lease_task

__all__ = ["HLDataSchemeAssessment"]

INITIAL_LEASE_TIME_SECONDS = 30 * 60
LEASE_TIME_UPDATE_DELTA_SECONDS = 30 * 60


class HLDataSchemeAssessment(aiko.DataScheme):

    def create_sources(self, stream, data_sources, frame_generator=None, use_create_frame=True):
        if use_create_frame:
            """
            - Not sure how to determine the frame_data keys ahead of time.
            - Not sure how this makes sense for DataSource Elements that
              produce many frames for one file, ie VideoFileRead
            """
            NotImplementedError()

        client = HLClient.get_client()

        def iter_tasks(data_sources):
            for data_source in data_sources:
                task_id = data_source.replace("hltask://", "")

                task = lease_task(
                    client,
                    task_id=task_id,
                    lease_sec=INITIAL_LEASE_TIME_SECONDS,
                    set_status_to="RUNNING",
                )

                assessments = [f"hlassessment://{s.id}" for s in task.case.latest_submission]

        stream.variables["source_paths_generator"] = iter_tasks(data_sources)
        rate, _ = self.pipeline_element.get_parameter("rate", default=None)
        rate = float(rate) if rate else None
        self.pipeline_element.create_frames(stream, frame_generator, rate=rate)

        return aiko.StreamEvent.OKAY, {}


aiko.DataScheme.add_data_scheme("hltask", HLDataSchemeAssessment)



"""
For video streams:
- Collect the inferred entities in memory until the end of the stream
- At the end of the stream, create an Avro file on S3 with the annotations and observations
- Finally create a submission pointing to the Avro file
For single image and text file streams:
- Create a submission storing the annotations and observations in the database
"""

from datetime import datetime
from typing import Dict, List, Optional, Tuple
from uuid import UUID

from pydantic import BaseModel

import highlighter as hl
from highlighter import DataFile, Entity
from highlighter.agent.capabilities import Capability, StreamEvent, StreamState
from highlighter.agent.capabilities.assessment_read import VIDEO
from highlighter.agent.utilities import (
    EntityAggregator,
    FileAvroEntityWriter,
    S3AvroEntityWriter,
)
from highlighter.client import HLClient


class CreateSubmissionPayload(BaseModel):
    errors: List[str]


class AssessmentWrite(Capability):
    def __init__(self, context):
        super().__init__(context)
        self.client = HLClient.get_client()
        self.entity_agg = EntityAggregator()
        self.debug = False

    def process_frame(self, stream, data_files: List[DataFile], **kwargs) -> Tuple[StreamEvent, dict]:
        debug, found = self.get_parameter("debug")
        self.debug = debug if found else False
        minimum_track_frame_length, found = self.get_parameter("minimum_track_frame_length")
        if not found:
            minimum_track_frame_length = 0
        self.entity_agg._minimum_track_frame_length = minimum_track_frame_length

        entities = {
            entity_id: entity for entities in kwargs.values() for entity_id, entity in entities.items()
        }

        data_file_ids = [df.file_id for df in data_files]
        stream_media_type = stream.variables["stream_media_type"]
        if stream_media_type != VIDEO:
            task_id = stream.stream_id  # hl_agent.py sets the stream ID to the task ID
            create_submission_from_entities(
                client=self.client,
                entities=entities,
                data_file_ids=data_file_ids,
                task_id=task_id,
            )
            self.logger.info(f"Created submission for task {task_id}")
        else:
            self.entity_agg.append_entities(list(entities.values()))
        self.logger.debug(f"ProcessFrame: {self.__class__.__name__}")

        return StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        now_str = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        task_id = stream_id  # hl_agent.py sets the stream ID to the task ID

        if stream.state == StreamState.ERROR:
            diagnostic = f"Stream '{stream_id}' stopped with error condition, not creating submission"
            self.logger.info(diagnostic)
            return StreamEvent.ERROR, {"diagnostic": diagnostic}

        if "stream_media_type" not in stream.variables:
            raise ValueError("The stream was not successfully created")

        if stream.variables["stream_media_type"] != VIDEO:
            pass
        elif self.debug:
            output_filename = f"{task_id}_{now_str}_DEBUG.avro"
            entity_writer = FileAvroEntityWriter(
                hl.ENTITY_AVRO_SCHEMA,
                output_filename,
            )
            entity_writer.write(self.entity_agg._track_entities)
        else:
            output_filename = f"{task_id}_{now_str}.avro"
            entity_writer = S3AvroEntityWriter(
                hl.ENTITY_AVRO_SCHEMA,
                output_filename,
                self.client,
            )

            shrine_file = entity_writer.write(self.entity_agg._track_entities)

            data_file_ids = stream.variables["video_data_file_ids"]
            create_submission_payload = self.client.createSubmission(
                return_type=CreateSubmissionPayload,
                status="completed",
                taskId=task_id,
                backgroundInfoLayerFileData=shrine_file,
                dataFileIds=[str(id) for id in data_file_ids],
            )
            if len(create_submission_payload.errors) > 0:
                diagnostic = str(create_submission_payload.errors)
                self.logger.error(diagnostic)
                return StreamEvent.ERROR, {"diagnostic": diagnostic}
            self.logger.info(f"Created submission for task {task_id} with backgroundInfoLayerFileData")
        return StreamEvent.OKAY, {}


def create_submission_from_entities(
    client: HLClient,
    entities: Dict[UUID, Entity],
    data_file_ids: List[UUID],
    task_id: Optional[UUID],
):
    annotations_attributes = []
    eavt_attributes = []
    for entity in entities.values():
        for annotation in entity.annotations:
            annotations_attributes.append(annotation.gql_dict())
            for eavt in annotation.observations:
                eavt_attributes.append({**eavt.gql_dict(), "annotationUuid": str(annotation.id)})
        for eavt in entity.global_observations:
            eavt_attributes.append(eavt.gql_dict())
    result = client.createSubmission(
        return_type=CreateSubmissionPayload,
        imageUuid=str(data_file_ids[0]),
        status="completed",
        taskId=str(task_id),
        annotationsAttributes=annotations_attributes,
        eavtAttributes=eavt_attributes,
        dataFileIds=[str(id) for id in data_file_ids],
    )
    if len(result.errors) > 0:
        raise ValueError(f"GraphQL Error: {result.errors}")



# To Do
# ~~~~~
# - Design refactor to disambiguate DataSource and DataScheme implementations
#   - In particular, aim to generalize "VideoReadFile,frame_generator()"
#
# - Move DataSource and DataTarget into "aiko_services.main.data_source.py" ?
#   - Define "data_sources" and "data_targets" as Python data classes
#
# - Improve "aiko_services/main/utilties/parser.py:parse()" to combine
#     "command" and "parameters" into a single parameter !

from aiko_services.main import *
from aiko_services.main.utilities import *

__all__ = ["DataSource", "DataTarget"]

# --------------------------------------------------------------------------- #
# DataSource: PipelineElement that loads frames of data from given locations
#
# Function arguments
# - use_create_frame: Enables using the more efficient (thread-less)
#     create_frame() method for a single path, rather than always using
#     create_frames() method and a "frame generator" (requires a thread)
#
# Parameters
# - data_sources: List of URLs that represent the locations to load data
# - data_batch_size: How many data items to be grouped per frame, default: 1
# - rate: How many frames to create per second, default: None (fast as possible)

# Each Pipeline Stream may have an individual DataSource DataScheme instance.
# Therefore, DataScheme instance variables are per-Stream variables :)

class DataSource(PipelineElementImpl):
    def _get_data_sources(self):
        data_sources, found = self.get_parameter("data_sources", required=True)
        data_sources = parse(data_sources, car_cdr=False)
        scheme = DataScheme.parse_data_url_scheme(data_sources[0])
        return data_sources, scheme

    def start_stream(self, stream, stream_id,
        frame_generator=None, use_create_frame=True):

        stream.variables["data_scheme"] = None

        try:
            data_sources, scheme = self._get_data_sources()
        except KeyError as key_error:
            return StreamEvent.ERROR, {"diagnostic": key_error}

        if not scheme in DataScheme.LOOKUP:
            diagnostic = f'DataSource URL scheme "{scheme}" is not supported'
            return StreamEvent.ERROR, {"diagnostic": diagnostic}

        data_scheme = DataScheme.LOOKUP[scheme](self)
        stream.variables["data_scheme"] = data_scheme

        return data_scheme.create_sources(stream, data_sources,
            frame_generator=frame_generator,
            use_create_frame=use_create_frame)

    def stop_stream(self, stream, stream_id):
        if "data_scheme" in stream.variables:
            stream.variables["data_scheme"].destroy_sources(stream)
            del stream.variables["data_scheme"]
        return StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
# DataTarget: PipelineElement that stores frames of data at given locations
#
# Parameters
# - data_targets: List of URLs that represent the locations to store data
#
# Each Pipeline Stream may have an individual DataTarget DataScheme instance.
# Therefore, DataScheme instance variables are per-Stream variables :)

class DataTarget(PipelineElementImpl):
    def _get_data_targets(self):
        data_targets, found = self.get_parameter("data_targets", required=True)
        data_targets = parse(data_targets, car_cdr=False)
        scheme = DataScheme.parse_data_url_scheme(data_targets[0])
        return data_targets, scheme

    def start_stream(self, stream, stream_id):
        stream.variables["data_scheme"] = None

        try:
            data_targets, scheme = self._get_data_targets()
        except KeyError as key_error:
            return StreamEvent.ERROR, {"diagnostic": key_error}

        if not scheme in DataScheme.LOOKUP:
            diagnostic = f'DataTarget URL scheme "{scheme}" is not supported'
            return StreamEvent.ERROR, {"diagnostic": diagnostic}

        data_scheme = DataScheme.LOOKUP[scheme](self)
        stream.variables["data_scheme"] = data_scheme

        return data_scheme.create_targets(stream, data_targets)

    def stop_stream(self, stream, stream_id):
        if "data_scheme" in stream.variables:
            stream.variables["data_scheme"].destroy_targets(stream)
            del stream.variables["data_scheme"]
        return StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



import os
from pathlib import Path

from .labeled_uuid import LabeledUUID

__all__ = [
    "DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE",
    "OBJECT_CLASS_ATTRIBUTE_UUID",
    "PIXEL_LOCATION_ATTRIBUTE_UUID",
    "DATA_FILE_ATTRIBUTE_UUID",
    "EMBEDDING_ATTRIBUTE_UUID",
    "TRACK_ATTRIBUTE_UUID",
    "HL_AGENT_QUEUE_RESPONSE_MAX_SIZE",
    "HL_CACHE_DIR",
    "HL_DOWNLOAD_TIMEOUT",
]

DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE = "DeprecatedCapabilityImplementationFile"

OBJECT_CLASS_ATTRIBUTE_UUID = LabeledUUID("df10b67d-b476-4c4d-acc2-c1deb5a0e4f4", label="object_class")

PIXEL_LOCATION_ATTRIBUTE_UUID = LabeledUUID("594fcdba-c3dc-4fad-b1c1-f5f537e1d16c", label="pixel_location")

DATA_FILE_ATTRIBUTE_UUID = LabeledUUID("2b13c2d0-02ae-4bcc-9086-df3cdacf3563", label="image")

EMBEDDING_ATTRIBUTE_UUID = LabeledUUID("8beab557-8d83-4257-82d0-101341236c5b", label="embedding")

TRACK_ATTRIBUTE_UUID = LabeledUUID("32a44d06-62be-4be8-9b0e-884af6dfd6f3", label="track")

HL_AGENT_QUEUE_RESPONSE_MAX_SIZE = int(os.environ.get("HL_AGENT_QUEUE_RESPONSE_MAX_SIZE", "100"))

HL_CACHE_DIR = Path(os.environ.get("HL_CACHE_DIR", Path.home() / ".highlighter-cache"))

HL_DOWNLOAD_TIMEOUT = int(os.environ.get("HL_DOWNLOAD_TIMEOUT", "300"))
